@@ -1,6 +1,6 @@
 /*
  * diff.c - diff specific util functions
- * Copyright (C) 2001, 2002, 2003, 2004 Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2001, 2002, 2003, 2004, 2005, 2009, 2011, 2013 Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  */
 
@@ -67,16 +67,20 @@ char *best_name (int n, char **names)
 {
         int *pathname_components;
         int *basename_length;
+	int *is_dev_null;
         int best_pn, best_bn, best_n, best = 0; /* shut gcc up */
         int i;
 
-        pathname_components = xmalloc (sizeof (int *) * n);
-        basename_length = xmalloc (sizeof (int *) * n);
+        pathname_components = xmalloc (sizeof (int) * n);
+        basename_length = xmalloc (sizeof (int) * n);
+        is_dev_null = xmalloc (sizeof (int) * n);
 
         best_pn = -1;
         for (i = 0; i < n; i++) {
-		if (!strcmp (names[i], "/dev/null"))
+		is_dev_null[i] = !strcmp (names[i], "/dev/null");
+		if (is_dev_null[i])
 			continue;
+
                 pathname_components[i] = num_pathname_components (names[i]);
                 if ((best_pn == -1) ||
                     (pathname_components[i] < best_pn))
@@ -87,7 +91,7 @@ char *best_name (int n, char **names)
         for (i = 0; i < n; i++) {
                 char *p;
 
-		if (!strcmp (names[i], "/dev/null"))
+		if (is_dev_null[i])
 			continue;
 
                 if (pathname_components[i] != best_pn)
@@ -109,6 +113,12 @@ char *best_name (int n, char **names)
 	for (i = 0; i < n; i++) {
 		int len;
 
+		if (is_dev_null[i])
+                        continue;
+
+                if (pathname_components[i] != best_pn)
+			continue;
+
                 if (basename_length[i] != best_bn)
                         continue;
 
@@ -122,6 +132,7 @@ char *best_name (int n, char **names)
 
         free (pathname_components);
         free (basename_length);
+        free (is_dev_null);
         return names[best];
 }
 
@@ -183,7 +194,7 @@ int read_atatline (const char *atatline,
 		if (!p)
 			return 1;
 		p++;
-		res = strtoul (p, &endptr, 0);
+		res = strtoul (p, &endptr, 10);
 		if (p == endptr)
 			return 1;
 		*orig_offset = res;
@@ -197,7 +208,7 @@ int read_atatline (const char *atatline,
 		if (!p)
 			return 1;
 		p++;
-		res = strtoul (p, &endptr, 0);
+		res = strtoul (p, &endptr, 10);
 		if (p == endptr)
 			return 1;
 		*new_offset = res;
@@ -287,7 +298,7 @@ static void copy_context_hunks (char **line, size_t *linelen, ssize_t *got,
 }
 
 static void convert_unified_hunks_to_context (char **line, size_t *linelen,
-					      unsigned long *linenum)
+					      ssize_t *got, unsigned long *linenum)
 {
 	unsigned long orig_offset, orig_count = 0, new_offset, new_count = 0;
 	char **orig_line = NULL, **new_line = NULL;
@@ -296,23 +307,25 @@ static void convert_unified_hunks_to_context (char **line, size_t *linelen,
 	unsigned int n_whats = 0;
 	unsigned long orig_linenum, new_linenum;
 	const char *no_newline_str = "\\ No newline at end of file\n";
+	char *misc = NULL;
 
 	if (feof (stdin))
 		goto eof;
 
-	if (getline (line, linelen, stdin) == -1)
+	if ((*got = getline (line, linelen, stdin)) == -1)
 		goto eof;
 	++*linenum;
 
 	for (;;) {
 		char *last_orig = NULL;
 		char *last_new = NULL;
-		char *misc = NULL;
 		char *what = NULL;
 		int newline = 1;
 		int can_omit_from = 1, can_omit_to = 1;
 		unsigned long i;
 
+		misc = NULL;
+
 		if (read_atatline (*line, &orig_offset, &orig_count,
 				   &new_offset, &new_count))
 			return;
@@ -332,7 +345,7 @@ static void convert_unified_hunks_to_context (char **line, size_t *linelen,
 		       (new_linenum < new_count) || newline) {
 			int get_out = 0;
 
-			if (getline (line, linelen, stdin) == -1)
+			if ((*got = getline (line, linelen, stdin)) == -1)
 				/* Should write out everything to date? */
 				break;
 			++*linenum;
@@ -344,6 +357,20 @@ static void convert_unified_hunks_to_context (char **line, size_t *linelen,
 
 			switch (**line) {
 			case ' ':
+				if (orig_linenum == orig_count)
+					/* We've already seen all the orig
+					 * lines we were expecting. */
+					error (EXIT_FAILURE, 0,
+					       "Garbled input at line %lu",
+					       *linenum);
+
+				if (new_linenum == new_count)
+					/* We've already seen all the new
+					 * lines we were expecting. */
+					error (EXIT_FAILURE, 0,
+					       "Garbled input at line %lu",
+					       *linenum);
+
 				what = NULL;
 				orig_what[orig_linenum] = " ";
 				new_what[new_linenum] = " ";
@@ -354,6 +381,13 @@ static void convert_unified_hunks_to_context (char **line, size_t *linelen,
 				break;
 
 			case '-':
+				if (orig_linenum == orig_count)
+					/* We've already seen all the orig
+					 * lines we were expecting. */
+					error (EXIT_FAILURE, 0,
+					       "Garbled input at line %lu",
+					       *linenum);
+
 				if (what) {
 					if (*what != '-')
 						*what = '!';
@@ -370,6 +404,13 @@ static void convert_unified_hunks_to_context (char **line, size_t *linelen,
 				break;
 
 			case '+':
+				if (new_linenum == new_count)
+					/* We've already seen all the new
+					 * lines we were expecting. */
+					error (EXIT_FAILURE, 0,
+					       "Garbled input at line %lu",
+					       *linenum);
+
 				if (what) {
 					if (*what != '+')
 						*what = '!';
@@ -533,7 +574,7 @@ static void do_convert_to_context (void)
 			fwrite (line + 4, (size_t) got - 4, 1, stdout);
 			puts ("***************");
 			convert_unified_hunks_to_context (&line, &linelen,
-							  &linenum);
+							  &got, &linenum);
 		}
 	}
 
@@ -583,7 +624,7 @@ static void copy_unified_hunks (char **line, size_t *linelen, ssize_t *got,
 }
 
 static void convert_context_hunks_to_unified (char **line, size_t *linelen,
-					      unsigned long *linenum)
+					      ssize_t *got, unsigned long *linenum)
 {
 	int happy = 1;
 
@@ -597,7 +638,6 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 		size_t *linelengths[2];
 		size_t n_lines[2];
 		size_t at[2];
-		ssize_t got;
 
 		n_lines[0] = n_lines[1] = 0;
 		linelengths[0] = linelengths[1] = NULL;
@@ -610,11 +650,13 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 			if (feof (stdin))
 				goto eof;
 
-			if (getline (line, linelen, stdin) == -1)
+			*got = getline (line, linelen, stdin);
+			if (*got == -1)
 				goto eof;
 			++*linenum;
 
-			if (!i && !strncmp (*line, "***************", 15)) {
+			if (!i && !misc &&
+			    !strncmp (*line, "***************", 15)) {
 				char *m = *line + 15;
 				if (strcmp (m, "\n"))
 					misc = xstrdup (m);
@@ -660,8 +702,8 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 			memset (lines[i], 0, sizeof (char *) * line_count[i]);
 
 			for (lnum = 0; lnum < line_count[i]; lnum++) {
-				got = getline (line, linelen, stdin);
-				if (got == -1)
+				*got = getline (line, linelen, stdin);
+				if (*got == -1)
 					goto eof;
 				++*linenum;
 
@@ -675,10 +717,10 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 					}
 				}
 
-				lines[i][lnum] = xmalloc ((size_t) got + 1);
-				memcpy (lines[i][lnum], *line, (size_t) got);
-				lines[i][lnum][got] = '\0';
-				linelengths[i][lnum] = (size_t) got;
+				lines[i][lnum] = xmalloc ((size_t) *got + 1);
+				memcpy (lines[i][lnum], *line, (size_t) *got);
+				lines[i][lnum][*got] = '\0';
+				linelengths[i][lnum] = (size_t) *got;
 				if (**line == ' ')
 					unchanged++;
 			}
@@ -699,6 +741,7 @@ static void convert_context_hunks_to_unified (char **line, size_t *linelen,
 		while (at[0] < n_lines[0] || at[1] < n_lines[1]) {
 			char *l[2] = { NULL, NULL };
 			size_t llen[2];
+			llen[0] = llen[1] = 0;
 			if (lines[0] && at[0] < n_lines[0]) {
 				l[0] = lines[0][at[0]];
 				llen[0] = linelengths[0][at[0]];
@@ -841,7 +884,7 @@ static void do_convert_to_unified (void)
 			printf ("+++ ");
 			fwrite (line + 4, (size_t) got - 4, 1, stdout);
 			convert_context_hunks_to_unified (&line, &linelen,
-							  &linenum);
+							  &got, &linenum);
 		} else {
 			fwrite (line, (size_t) got, 1, stdout);
 			got = getline (&line, &linelen, stdin);
@@ -892,9 +935,11 @@ static FILE *do_convert (FILE *f, const char *mode, int seekable,
 			/* Parent. */
 			close (fildes[1]);
 			ret = fdopen (fildes[0], mode);
+			if (!ret)
+			    error (EXIT_FAILURE, errno, "fdopen failed");
 
 			if (seekable) {
-				FILE *tmp = tmpfile ();	
+				FILE *tmp = xtmpfile ();	
 				while (!feof (ret)) {
 					int c = fgetc (ret);
 
@@ -904,6 +949,7 @@ static FILE *do_convert (FILE *f, const char *mode, int seekable,
 					fputc (c, tmp);
 				}
 
+				fclose (ret);
 				rewind (tmp);
 				return tmp;
 			}
