@@ -1,6 +1,6 @@
 /*
  * rediff - fix offset and counts of a hand-edited diff
- * Copyright (C) 2001, 2002, 2004 Tim Waugh <twaugh@redhat.com>
+ * Copyright (C) 2001, 2002, 2004, 2009, 2011, 2019 Tim Waugh <twaugh@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  *
  */
 
@@ -127,6 +127,7 @@ static unsigned long adjust_offsets_and_copy (long *offset, FILE *in,
 	size_t linelen = 0;
 	unsigned long orig_offset, orig_lines, new_offset, new_lines;
 	unsigned long count = 0;
+	char *trailing;
 
 	if (getline (&line, &linelen, in) == -1)
 		goto out;
@@ -154,7 +155,12 @@ static unsigned long adjust_offsets_and_copy (long *offset, FILE *in,
 		error (EXIT_FAILURE, 0, "Line not understood: %s", line);
 	}
 
-	free (line);
+	/* Find the part after "@@...@@". */
+	trailing = strchr (line, '+');
+	trailing += strcspn (trailing, " \n");
+	if (*trailing == ' ')
+		trailing++;
+	trailing += strspn (trailing, "@");
 
 	/* Adjust offsets */
 	fprintf (out, "@@ -%lu", orig_offset);
@@ -163,7 +169,8 @@ static unsigned long adjust_offsets_and_copy (long *offset, FILE *in,
 	fprintf (out, " +%lu", new_offset + *offset);
 	if (new_lines != 1)
 		fprintf (out, ",%lu", new_lines);
-	fprintf (out, " @@\n");
+	fprintf (out, " @@%s", trailing);
+	free (line);
 
 	/* Copy remaining lines of hunk */
 	count += copy_hunk (in, out, orig_lines, new_lines);
@@ -250,7 +257,7 @@ static long added_hunk (const char *meta, long offset, FILE *modify, FILE *t,
 	char *p = strchr (meta, '-'), *q;
 	unsigned long orig_offset = 0, new_offset;
 	unsigned long orig_count = 0, new_count = 0;
-	FILE *newhunk = tmpfile ();
+	FILE *newhunk = xtmpfile ();
 
 	if (!newhunk)
 		error (EXIT_FAILURE, errno, "Couldn't create temporary file");
@@ -261,7 +268,8 @@ static long added_hunk (const char *meta, long offset, FILE *modify, FILE *t,
 	}
 
 	if (!p || p == q) {
-		p[strlen (p) - 1] = '\0';
+		if (p)
+			p[strlen (p) - 1] = '\0';
 		error (EXIT_FAILURE, 0,
 		       "Hunk addition requires original line: %s", meta);
 	}
@@ -344,6 +352,7 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 	if (line[0] == '+' && line[1] == '@') {
 		/* Minimally correct modified @@ banners. */
 		unsigned long oo, no;
+		char *trailing;
 		if (read_atatline (line + 1, &oo, NULL, &no, NULL))
 			goto out;
 
@@ -354,6 +363,12 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 			hunk->info->info_written = 1;
 		}
 
+		trailing = strchr (line + 1, '+');
+		trailing += strcspn (trailing, " \n");
+		if (*trailing == ' ')
+			trailing++;
+		trailing += strspn (trailing, "@");
+
 		if (oo != orig_offset)
 			no = new_offset + oo - orig_offset;
 		else
@@ -364,7 +379,7 @@ static long removed_hunk (const char *meta, FILE *modify, FILE *t,
 		fprintf (t, " +%lu", no);
 		if (new_count != 1)
 			fprintf (t, ",%lu", new_count);
-		fprintf (t, " @@\n");
+		fprintf (t, " @@%s", trailing);
 		goto out;
 	}
 
@@ -451,10 +466,11 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 	unsigned long morig_offset, morig_count, mnew_offset, mnew_count;
 	char *line = NULL;
 	size_t linelen = 0;
-	FILE *t = tmpfile ();
+	FILE *t = xtmpfile ();
 	int t_written_to = 0;
 	unsigned long i, at = 1;
 	unsigned long replaced, unaltered;
+	char *trailing;
 	int r;
 
 	if (!t)
@@ -479,6 +495,13 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 	calc_new_offset = new_offset;
 	calc_new_count = new_count;
 
+	trailing = strchr (line, '+');
+	trailing += strcspn (trailing, " \n");
+	if (*trailing == ' ')
+		trailing++;
+	trailing += strspn (trailing, "@");
+	trailing = xstrdup (trailing);
+
 	r = getline (&line, &linelen, modify);
 	assert (r != -1);
 	r = read_atatline (line, &morig_offset, &morig_count,
@@ -488,7 +511,10 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 
 	do {
 		/* Lines before the modification are unaltered. */
-		int trim = 0;
+                int trim = 0, skip_trim;
+		if (morig_offset < hunk->line_in_diff)
+			error (EXIT_FAILURE, errno, "Invalid changes made");
+
 		unaltered = morig_offset - hunk->line_in_diff;
 		if (!morig_count)
 			unaltered++;
@@ -521,6 +547,7 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 		}
 
 		while (morig_count || mnew_count) {
+                        skip_trim = 0;
 			if (getline (&line, &linelen, modify) == -1)
 				break;
 
@@ -538,7 +565,7 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 				case '+':
 					this_offset--;
 					calc_new_count--;
-					trim = 0;
+					skip_trim = 1;
 					break;
 				case '-':
 					this_offset++;
@@ -556,7 +583,7 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 					       line[0], line[1]);
 				}
 
-				if (trim) {
+				if (trim && !skip_trim) {
 					orig_offset++;
 					calc_new_offset++;
 				}
@@ -745,9 +772,10 @@ static long show_modified_hunk (struct hunk **hunkp, long line_offset,
 		fprintf (out, " +%lu", calc_new_offset + line_offset);
 		if (calc_new_count != 1)
 			fprintf (out, ",%lu", calc_new_count);
-		fprintf (out, " @@\n");
+		fprintf (out, " @@%s", trailing);
 	}
 
+	free (trailing);
 	while (getline (&line, &linelen, t) != -1)
 		fputs (line, out);
 
@@ -782,7 +810,6 @@ static int rediff (const char *original, const char *edited, FILE *out)
 	struct hunk *hunks = NULL, **p = &hunks, *last = NULL;
 	struct hunk *current_hunk = NULL;
 	long line_offset = 0;
-	fpos_t first_hunk;
 
 	/* Let's take a look at what hunks are in the original diff. */
 	o = xopen (original, "rbm");
@@ -880,7 +907,7 @@ static int rediff (const char *original, const char *edited, FILE *out)
 	/* Run diff between original and edited. */
 	t = xpipe (DIFF, &child, "r", DIFF, "-U0",
 		   original, edited, NULL);
-	m = tmpfile ();
+	m = xtmpfile ();
 	if (m) {
 		size_t buffer_size = 10000;
 		char *buffer = xmalloc (buffer_size);
@@ -962,8 +989,7 @@ static int rediff (const char *original, const char *edited, FILE *out)
 
 			/* This meta hunk is the first pertaining to
 			 * the hunk in the original. */
-			first_hunk = pos;
-			t = tmpfile ();
+			t = xtmpfile ();
 		}
 
 		current_hunk = which;
@@ -1042,8 +1068,8 @@ int main (int argc, char *argv[])
 		char *p = xmalloc (strlen (argv[0]) +
 				   strlen ("recountdiff") + 1);
 		char *f;
-		char **const new_argv = xmalloc (sizeof (char *) * argc);
-		memcpy (new_argv, argv, sizeof (char *) * argc);
+		char **const new_argv = xmalloc (sizeof (char *) * (argc + 1));
+		memcpy (new_argv, argv, sizeof (char *) * (argc + 1));
 		new_argv[0] = p;
 		strcpy (p, argv[0]);
 		f = strrchr (p, '/');
